\begin{snippetimport}
\begin{code}
import qualified Data.List as List
\end{code}
\end{snippetimport}

\begin{snippetortholattice}
\begin{code}
infixr 5 /\, \/   -- $\wedge$ and $\vee$ are right associative with precedence $5$

class Ortholattice t where
  (/\) :: t -> t -> t      -- meet
  (\/) :: t -> t -> t      -- join
  cpl  :: t -> t           -- orthocomplement
  zero :: t                -- least element
  unit :: t                -- greatest element

  a /\ b = cpl (cpl a \/ cpl b)
  a \/ b = cpl (cpl a /\ cpl b)
  zero   = cpl unit
  unit   = cpl zero
\end{code}
\end{snippetortholattice}

\begin{snippetatomicops}
\begin{code}
mo2l   :: (Ortholattice t) => t -> t -> t
mo2r   :: (Ortholattice t) => t -> t -> t
mo2l'  :: (Ortholattice t) => t -> t -> t
mo2r'  :: (Ortholattice t) => t -> t -> t
upcom  :: (Ortholattice t) => t -> t -> t

a `mo2l`   b  = let a' = cpl a; b' = cpl b in a  /\ (a'\/b ) /\ (a'\/b')
a `mo2r`   b  = let a' = cpl a; b' = cpl b in b  /\ (a \/b') /\ (a'\/b')
a `mo2l'`  b  = let a' = cpl a; b' = cpl b in a' /\ (a \/b ) /\ (a \/b')
a `mo2r'`  b  = let a' = cpl a; b' = cpl b in b' /\ (a \/b ) /\ (a'\/b )
a `upcom`  b  = let a' = cpl a; b' = cpl b in (a\/b) /\ (a\/b') /\ (a'\/b) /\ (a'\/b')
\end{code}
\end{snippetatomicops}

\begin{snippetmotwo}
\begin{code}
data MO2 = O | A | B | B' | A' | I  deriving (Eq, Enum, Bounded)

instance Ortholattice MO2 where

  O  /\ _           = O    -- $\mathbf{0} \wedge x = \mathbf{0}$
  _  /\ O           = O    -- $x \wedge \mathbf{0} = \mathbf{0}$
  I  /\ x           = x    -- $\mathbf{1} \wedge x = x$
  x  /\ I           = x    -- $x \wedge \mathbf{1} = x$
  x  /\ y | x == y  = x    -- $x \wedge x = x$
  _  /\ _           = O    -- $x \wedge y = \mathbf{0}$ otherwise

  I  \/ _           = I    -- $\mathbf{1} \vee x = \mathbf{1}$
  _  \/ I           = I    -- $x \vee \mathbf{1} = \mathbf{1}$
  O  \/ x           = x    -- $\mathbf{0} \vee x = x$
  x  \/ O           = x    -- $x \vee \mathbf{0} = x$
  x  \/ y | x == y  = x    -- $x \vee x = x$
  _  \/ _           = I    -- $x \vee y = \mathbf{1}$ otherwise

  cpl O   = I
  cpl A   = A'
  cpl B   = B'
  cpl B'  = B
  cpl A'  = A
  cpl I   = O

  zero = O
  unit = I
\end{code}
\end{snippetmotwo}

\begin{snippetbool}
\begin{code}
instance Ortholattice Bool where
  (/\) = (&&)    -- logical conjunction
  (\/) = (||)    -- logical disjunction
  cpl  = not     -- logical negation
  zero = False
  unit = True
\end{code}
\end{snippetbool}

\begin{snippetfoml}
\begin{code}
data F2OML = F2 MO2 Bool Bool Bool Bool

instance Ortholattice F2OML where

  F2 a al au ad ar /\ F2 b bl bu bd br = F2 (a/\b) (al/\bl) (au/\bu) (ad/\bd) (ar/\br)
  F2 a al au ad ar \/ F2 b bl bu bd br = F2 (a\/b) (al\/bl) (au\/bu) (ad\/bd) (ar\/br)

  cpl (F2 x xl xu xd xr) = F2 (cpl x) (cpl xl) (cpl xu) (cpl xd) (cpl xr)

  zero = F2 zero zero zero zero zero
  unit = F2 unit unit unit unit unit
\end{code}
\end{snippetfoml}

\begin{snippetcomfoml}
\begin{code}
comF2OML :: F2OML -> F2OML -> Bool
(F2 a _ _ _ _) `comF2OML` (F2 b _ _ _ _)  -- extract just the $\mathrm{MO}_2$ parts
 | a == zero   = True     -- $\mathbf{0} C b$
 | b == zero   = True     -- $a C \mathbf{0}$
 | a == unit   = True     -- $\mathbf{1} C b$
 | b == unit   = True     -- $a C \mathbf{1}$
 | a == b      = True     -- $a C a$
 | a == cpl b  = True     -- $a C a'$
 | otherwise   = False    -- $\neg(a C b)$ otherwise
\end{code}
\end{snippetcomfoml}

\begin{snippetprojectmotwo}
\begin{code}
projectMO2 :: F2OML -> F2OML
projectMO2 (F2 x _ _ _ _) = F2 x zero zero zero zero
\end{code}
\end{snippetprojectmotwo}

\begin{snippetevaluatemotwo}
\begin{code}
evaluateMO2 :: F2OML -> MO2 -> F2OML -> F2OML
evaluateMO2 a   _    b | a `comF2OML` b  = zero    -- if $a C b$, the result is always zero
evaluateMO2 _   O    _                   = zero
evaluateMO2 a   A    _                   = projectMO2 a          -- $a\ \Fa\ b$
evaluateMO2 _   B    b                   = projectMO2 b          -- $a\ \Fb\ b$
evaluateMO2 _   B'   b                   = projectMO2 (cpl b)    -- $a\ \FB\ b$
evaluateMO2 a   A'   _                   = projectMO2 (cpl a)    -- $a\ \FA\ b$
evaluateMO2 _   I    _                   = projectMO2 unit       -- $a\ \Fi\ b$
\end{code}
\end{snippetevaluatemotwo}

\begin{snippettriplex}
\begin{code}
evaluateTriplex :: F2OML -> F2OML -> F2OML -> F2OML
evaluateTriplex a (F2 f fl fu fd fr) b = let
  a'  = cpl a
  b'  = cpl b
  xl  = if fl  then a /\b'  else zero    -- $a\ \Fol\ b = a\phpr \wedge b'$
  xu  = if fu  then a'/\b'  else zero    -- $a\ \Fou\ b = a'     \wedge b'$
  xd  = if fd  then a /\b   else zero    -- $a\ \Fod\ b = a\phpr \wedge b$
  xr  = if fr  then a'/\b   else zero    -- $a\ \For\ b = a'     \wedge b$
  x   = evaluateMO2 a f b                -- evaluate the $\mathrm{MO}_2$ part
  in x \/ xl \/ xu \/ xd \/ xr           -- take the join of all results
\end{code}
\end{snippettriplex}

\begin{snippetmotwoshowread}
\begin{code}
instance Show MO2 where
  show O  = "0"
  show A  = "a"
  show A' = "A"
  show B  = "b"
  show B' = "B"
  show I  = "1"


instance Read MO2 where
  readsPrec _ "" = []
  readsPrec _ (c:ss)
   | c == '0'  = [(O,  ss)]
   | c == 'a'  = [(A,  ss)]
   | c == 'A'  = [(A', ss)]
   | c == 'b'  = [(B,  ss)]
   | c == 'B'  = [(B', ss)]
   | c == '1'  = [(I,  ss)]
   | otherwise = []
\end{code}
\end{snippetmotwoshowread}

\begin{snippetfomlshowread}
\begin{code}
instance Show F2OML where
  show (F2 x xl xu xd xr) = let
    sl = if xl then "l" else ""
    su = if xu then "u" else ""
    sd = if xd then "d" else ""
    sr = if xr then "r" else ""
    in show x ++ sl ++ su ++ sd ++ sr


instance Read F2OML where
  readsPrec _ "" = []
  readsPrec _ (s:sl) = let
    x = read [s] :: MO2
    (su, xl) = stripIfFound sl 'l'
    (sd, xu) = stripIfFound su 'u'
    (sr, xd) = stripIfFound sd 'd'
    (ss, xr) = stripIfFound sr 'r'
    in [(F2 x xl xu xd xr, ss)] where
      stripIfFound :: String -> Char -> (String, Bool)
      stripIfFound "" _                    = ("", False)
      stripIfFound s@(c:ss) c' | c == c'   = (ss, True)
                               | otherwise = (s,  False)
\end{code}
\end{snippetfomlshowread}

\begin{snippettable}
\begin{code}
drawTable :: (MO2 -> MO2 -> MO2) -> String
drawTable f = drawGrid where
  allOfMO2 :: [MO2]
  allOfMO2 = [minBound .. maxBound]
  drawElem :: (Show a) => a -> String
  drawElem i = " " ++ show i ++ " "
  drawLine :: MO2 -> String
  drawLine i = drawElem i ++ "\x2502" ++ List.concat (List.map (drawElem . (\j -> f i j)) allOfMO2)
  drawHead :: String
  drawHead = "   " ++ "\x2502" ++ List.concat (List.map drawElem allOfMO2)
  drawRule :: String
  drawRule = "\x2500\x2500\x2500\x253C" ++ List.concat (List.map (\_ -> "\x2500\x2500\x2500") allOfMO2)
  drawGrid :: String
  drawGrid = drawHead ++ "\n" ++ drawRule ++ "\n" ++ List.intercalate "\n   \x2502\n" (List.map drawLine allOfMO2)
\end{code}
\end{snippettable}

\begin{snippettriplexshowread}
\begin{code}
triplex :: String -> String -> String -> String
triplex a f b = show (evaluateTriplex (read a) (read f) (read b))
\end{code}
\end{snippettriplexshowread}

\begin{snippetmain}
\begin{code}

main :: IO ()
main = do
  putStrLn (triplex "Au" "ald" "blur")
  putStrLn (triplex "alur" "bdr" "Bu")
  putStrLn (triplex "Budr" "0d" "Aldr")
  putStrLn (triplex "Bl" "1ldr" "Au")
  putStrLn (triplex "Blr" "0" "aur")
  putStrLn (triplex "bld" "1ludr" "Aud")
  putStrLn (triplex "1" "1lr" "1")
  putStrLn (triplex "bdr" "0lur" "ald")

\end{code}
\end{snippetmain}